#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

Samller Utility dialogs not depending on app model   

"""

from PyQt6.QtWidgets                    import QLayout, QDialogButtonBox, QPushButton, QDialogButtonBox
from PyQt6.QtWidgets                    import QFileDialog, QWidget

from ..base.widgets                     import * 
from ..base.panels                      import Dialog

from ..model.airfoil                    import Airfoil, Flap_Definition
from ..model.polar_set                  import *

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class Airfoil_Save_Dialog (Dialog):
    """ 
    Common airfoil save dialog - optionally
        - set new filename
        - set new airfoil name
        - select directory
        - remove temp files 

    When user successfully selected an airfoil file, 'set' is called with 
    the new <Airfoil> as argument 
    """

    _width  = (520, None)
    _height = 300

    name = "Save Airfoil as..."

    def __init__ (self,*args, remove_designs=False, rename_mode=False, **kwargs):

        self._rename_mode    = rename_mode
        self._remove_designs = remove_designs

        if rename_mode:
            self.name = "Rename Airfoil"

        super().__init__ (*args, **kwargs)


    @property
    def airfoil (self) -> Airfoil:
        return self.dataObject_copy

    @property 
    def remove_designs (self) -> bool:
        """ remove designs and dir upon finish"""
        return self._remove_designs
    
    def set_remove_designs (self, aBool : bool):
        self._remove_designs = aBool


    def _init_layout(self) -> QLayout:

        l = QGridLayout()
        r = 0 
        Label  (l,r,0, colSpan=4, get="Change airfoil name and/or filename before saving the airfoil",
                       style=style.COMMENT )
        r += 1
        SpaceR (l, r, stretch=0) 
        r += 1
        Field  (l,r,0, lab="Name", obj= self.airfoil, prop=Airfoil.name, width=(150,None),
                       style=self._style_names, signal=True)
        Button (l,r,2, text="Use Filename", set=self.airfoil.set_name_from_fileName, width=90,
                       hide=self._names_are_equal, signal=True,
                       toolTip="Use filename as airfoil name")
        r += 1
        Label  (l,r,1, colSpan=4, get=self._messageText, style=style.COMMENT, height=20,
                       hide= lambda: not self._messageText())
        r += 1
        Field  (l,r,0, lab="Filename", obj=self.airfoil, prop=Airfoil.fileName, width=(150,None),
                signal=True)
        Button (l,r,2, text="Use Name", set=self.airfoil.set_fileName_from_name, width=90,
                       hide=self._names_are_equal, signal=True,
                       toolTip="Use airfoil name as filename")
        r += 1
        SpaceR (l, r, height=10, stretch=0) 
        r += 1
        Field  (l,r,0, lab="Directory", obj=self.airfoil, prop=Airfoil.pathName_abs, width=(150,None),
                       toolTip=lambda: f"{self.airfoil.pathName_abs}",
                       disable=True)
        ToolButton (l,r,2, icon=Icon.OPEN, set=self._open_dir, signal=True,
                    hide = self._rename_mode,
                    toolTip = 'Select directory of airfoil') 
        r += 1
        SpaceR (l, r, height=10, stretch=2) 
        r += 1
        CheckBox (l,r,0, text="Remove all designs and design directory", colSpan=4,
                        get=lambda: self.remove_designs, set=self.set_remove_designs)
        r += 1
        SpaceR (l, r, height=10, stretch=1) 

        l.setColumnStretch (1,5)
        l.setColumnMinimumWidth (0,80)
        l.setColumnMinimumWidth (2,35)

        return l


    def _on_widget_changed (self, *_):
        """ slot for change of widgets"""

        # delayed refresh as pressed button hides itself 
        timer = QTimer()                                
        timer.singleShot(20, self.refresh)     # delayed emit 


    def _names_are_equal (self) -> bool: 
        """ is airfoil name different from filename?"""
        fileName_without = os.path.splitext(self.airfoil.fileName)[0]
        return fileName_without == self.airfoil.name


    def _style_names (self):
        """ returns style.WARNING if names are different"""
        if not self._names_are_equal(): 
            return style.WARNING
        else: 
            return style.NORMAL


    def _messageText (self): 
        """ info / wanrning text"""
        if not self._names_are_equal():
             return "You may want to sync airfoil Name and Filename"


    def _open_dir (self):
        """ open directory and set to airfoil """

        select_dir = self.airfoil.pathName_abs     
        pathName_new = QFileDialog.getExistingDirectory(self, caption="Select directory",
                                                           directory=select_dir)
        if pathName_new:                         # user pressed open
            self.airfoil.set_pathName (pathName_new)



    def accept(self):
        """ Qt overloaded - ok - save airfoil"""

        # set original airfoil with data 
        airfoil : Airfoil = self.dataObject 
        airfoil_copy : Airfoil = self.dataObject_copy

        airfoil.set_name (airfoil_copy.name)
        airfoil.set_pathFileName (airfoil_copy.pathFileName, noCheck=True)

        airfoil.save ()

        super().accept() 



class Polar_Definition_Dialog (Dialog):
    """ Dialog to edit a single polar definition"""

    _width  = 480
    _height = (300, None)

    name = "Edit Polar Definition"

    def __init__ (self, parent : QWidget, polar_def : Polar_Definition, 
                  small_mode = False,                                       # with flap etc 
                  polar_type_fixed = False,                                 # change of polar type not allowed 
                  fixed_chord : float = None,                               # fixed chord length in mm
                  **kwargs): 

        self._polar_def         = polar_def
        self._small_mode        = small_mode
        self._polar_type_fixed  = polar_type_fixed
        self._fixed_chord       = fixed_chord
        self._has_xtrip         = polar_def.has_xtrip

        if small_mode:
            self._height = 140
            self._width  = 480

        # init layout etc 
        super().__init__ (parent=parent, **kwargs)

        if fixed_chord:
            self.setWindowTitle (f"{self.name} at Chord {fixed_chord}mm")


    @property
    def polar_def (self) -> Polar_Definition:
        return self._polar_def

    @property
    def flap_def (self) -> Flap_Definition:
        return self.polar_def.flap_def if self.polar_def else None

    @property
    def has_xtrip (self) -> bool:
        """ True if forced transition is or shall be set on top or bottom side"""
        return self._has_xtrip
    
    def set_has_xtrip (self, aBool : bool):
        self._has_xtrip = aBool
        if not aBool:
            self.polar_def.set_xtript (None)
            self.polar_def.set_xtripb (None)
        self.refresh()


    @property
    def v (self) -> float:
        """ velocity in m/s if chord is given"""
        if self._fixed_chord and self.polar_def:
            return v_from_re(self.polar_def.re, self._fixed_chord/1000, round_dec=None)
        else:
            return None

    def _init_layout(self) -> QLayout:

        l = QGridLayout()
        l.setVerticalSpacing(3)                                     # strange - otherwise vertical spacing is too large
        r,c = 0,0 
        SpaceR (l, r, height=10,stretch=0) 
        r += 1 
        Label  (l,r,c, get="Polar type")
        ComboBox (l,r,c+1,  width=55, options=polarType.values(),
                        obj=self.polar_def, prop=Polar_Definition.type,
                        disable=self._polar_type_fixed)
        r += 1
        FieldF (l,r,c, width=70, step=10, lim=(1, 99999), unit="k", dec=0,
                        lab=lambda: "Re number" if self.polar_def.type == polarType.T1 else "Re · √Cl", 
                        obj=self.polar_def, prop=Polar_Definition.re_asK)
        l.setColumnMinimumWidth (c,80)
        c += 2
        ToolButton  (l,r,c, icon=Icon.EDIT, set=self.calc_re,
                        toolTip="Calculate Re from velocity and chord length")
        c += 1
        SpaceC  (l,c, width=10)
        c += 1
        FieldF (l,r,c, lab="Mach", width=60, step=0.1, lim=(0, 1.0), dec=2,
                        obj=self.polar_def, prop=Polar_Definition.ma)
        l.setColumnMinimumWidth (c,45)
        c += 2
        SpaceC  (l,c, width=10)
        c += 1
        FieldF (l,r,c, lab="Ncrit", width=60, step=1, lim=(1, 20), dec=1,
                        obj=self.polar_def, prop=Polar_Definition.ncrit)
        l.setColumnMinimumWidth (c,45)
        c += 2
        SpaceC  (l,c, width=10, stretch=5)

        c = 0 
        
        if not self._small_mode:
            r += 1
            FieldF (l,r,c, width=70, unit="m/s", dec=1,
                            lab=lambda: "Velocity", obj=self, prop=Polar_Definition_Dialog.v,
                            disable=True,
                            hide = lambda: self._fixed_chord is None or self.polar_def.type==polarType.T2)
            r += 1 
            SpaceR (l, r, height=10, stretch=0) 

            r += 1 
            CheckBox (l,r,c, text="Force transition at x ...", colSpan=7,
                            obj=self, prop=Polar_Definition_Dialog.has_xtrip)
            r += 1
            FieldF  (l,r,c, lab="Upper Side", width=60, step=1, lim=(0.0, 100), dec=0, unit='%',
                            obj=self.polar_def, prop=Polar_Definition.xtript,
                            hide=lambda: not self.has_xtrip)
            FieldF  (l,r,c+4, lab="Lower", width=60, step=1, lim=(0.0, 100), dec=0, unit='%',
                            obj=self.polar_def, prop=Polar_Definition.xtripb,
                            hide=lambda: not self.has_xtrip)
            r += 1 
            SpaceR (l, r, height=5, stretch=1) 

            r += 1 
            CheckBox (l,r,c, text="Set flap just for this polar ...", colSpan=7,
                            obj=self.polar_def, prop=Polar_Definition.is_flapped)
            r += 1
            FieldF  (l,r,c, lab="Flap Angle", width=60, step=0.1, lim=(-20,20), dec=1, unit='°', 
                            obj=lambda: self.flap_def, prop=Flap_Definition.flap_angle,
                            hide=lambda: not self.polar_def.is_flapped)
            FieldF  (l,r,c+4, lab="Hinge x", width=60, step=1, lim=(1, 98), dec=1, unit="%",
                            obj=lambda: self.flap_def, prop=Flap_Definition.x_flap,
                            hide=lambda: not self.polar_def.is_flapped)
            FieldF  (l,r,c+7, lab="Hinge y", width=60, step=1, lim=(0, 100), dec=0, unit='%',
                            obj=lambda: self.flap_def, prop=Flap_Definition.y_flap,
                            hide=lambda: not self.polar_def.is_flapped)
            # Label   (l,r,c+10, get="of thickness", style=style.COMMENT)

            r += 1
            SpaceR (l, r, height=5, stretch=1) 
            r += 1 
            CheckBox (l,r,c, text=lambda: f"Auto Range of polar {self.polar_def.specVar} values for a complete polar", colSpan=7,
                            get=self.polar_def.autoRange)
            r += 1
            FieldF (l,r,c, lab=f"Step {var.ALPHA}", width=60, step=0.1, lim=(0.1, 1.0), dec=2,
                            obj=self.polar_def, prop=Polar_Definition.valRange_step,
                            hide = lambda: self.polar_def.specVar != var.ALPHA)
            FieldF (l,r,c, lab=f"Step {var.CL}", width=60, step=0.01, lim=(0.01, 0.1), dec=2,
                            obj=self.polar_def, prop=Polar_Definition.valRange_step,
                            hide = lambda: self.polar_def.specVar != var.CL)
            Label  (l,r,c+3, style=style.COMMENT, colSpan=6, 
                            get="The smaller the value, the more time is needed")
        r += 1
        l.setRowStretch (r,1)
        return l


    def calc_re (self):
        """ calc re from velocity and chord length"""

        if self.polar_def.type == polarType.T1:
            dialog = Calc_Reynolds_Dialog (self, re_asK=self.polar_def.re_asK, fixed_chord=self._fixed_chord,
                                        parentPos=(0.5, 0.5), dialogPos=(0,1))
        else:
            dialog = Calc_Re_Sqrt_Cl_Dialog (self, re_asK=self.polar_def.re_asK, fixed_chord=self._fixed_chord,
                                        parentPos=(0.5, 0.5), dialogPos=(0,1))

        dialog.exec()     

        if dialog.has_been_set:
            self.polar_def.set_re_asK (dialog.re_asK)
            self.refresh()


    @override
    def _on_widget_changed (self):
        """ slot a input field changed - repanel and refresh"""
        self.refresh()
        # Adjust dialog size to fit content after widgets are shown/hidden
        self.adjustSize()


    @override
    def _button_box (self):
        """ returns the QButtonBox with the buttons of self"""

        buttons = QDialogButtonBox.StandardButton.Close
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.rejected.connect(self.close)

        return buttonBox 


    @override
    def reject(self): 
        """ close or x-Button pressed"""

        # ensure no flap def with flap angle == 0.0 
        if self.flap_def and self.flap_def.flap_angle == 0.0:
            self.polar_def.set_is_flapped (False) 

        # normal close 
        super().reject()



class Airfoil_Scale_Dialog (Dialog):
    """ small dialog to edit scale factor of an (reference) airfoil"""

    _width  = 320
    _height = 170

    name = "Set Scale Value"

    def __init__ (self, *args, **kwargs): 

        self._close_btn  : QPushButton = None 

        super().__init__ ( *args, **kwargs)

        self._close_btn.clicked.connect  (self.close)


    @property
    def scale_factor (self) -> float:
        return self.dataObject_copy

    def set_scale_factor (self, aVal):
        self._dataObject_copy = aVal


    def _init_layout(self) -> QLayout:

        l = QGridLayout()
        r,c = 0,0 
        Label  (l,r,c, style=style.COMMENT, height=80, colSpan=3,
                get="Set a scale value for the selected reference airfoil.<br>" +\
                    "This also scales the Reynolds number of its polars<br>" +\
                    "allowing to compare airfoils at their wing section.<br>")
        r += 1
        FieldF (l,r,c, lab="Scale to", width=60, unit="%", step=10, dec=0, lim=(5,500),
                obj=self, prop=Airfoil_Scale_Dialog.scale_factor)
        r += 1

        l.setRowStretch (r,1)    
        l.setColumnMinimumWidth (0,80)
        l.setColumnStretch (2,2)

        return l


    @override
    def _button_box (self):
        """ returns the QButtonBox with the buttons of self"""

        buttonBox = QDialogButtonBox (QDialogButtonBox.StandardButton.Close) #  | QDialogButtonBox.StandardButton.Cancel)
        self._close_btn  = buttonBox.button(QDialogButtonBox.StandardButton.Close)
        return buttonBox 



class Calc_Reynolds_Dialog (Dialog):
    """ Little dialog to calculate Reynolds from velocity and chord"""

    _width  = 310
    _height = 200

    name = "Calculate Reynolds Number"


    def __init__ (self, parent : QWidget,  
                  re_asK : int = None,                  # initial Reynolds in 1000
                  fixed_chord : float = None,           # fixed chord length in mm
                  **kwargs): 

       
        self._has_been_set = False
        self._v      = 30.0
        self._chord  = fixed_chord if fixed_chord is not None else 200
        self._chord_fixed = fixed_chord is not None

        super().__init__ (parent, **kwargs)
        
        if re_asK is not None:
            self.set_re_asK (re_asK)                # will also set v
            self._has_been_set = False              # but not yet confirmed by user


    @property
    def re_asK (self) -> int: 
        """ Reynolds number base 1000"""
        return int(re_from_v(self.v, self.chord/1000, round_to=None) / 1000)
     
    def set_re_asK (self, aVal : float): 

        if not isinstance(aVal, (int, float)):
            return
        self.set_v (v_from_re(aVal*1000, self.chord/1000, round_dec=None))              # chord in m


    @property
    def v (self) -> float:
        """ velocity in m/s """
        return self._v  
    
    def set_v (self, aVal : float):
        if not isinstance(aVal, (int, float)):
            return
        self._v = aVal
        self._has_been_set = True
        self.refresh()


    @property
    def chord (self) -> float:
        """ chord in mm """
        return self._chord
    def set_chord (self, aVal : float):
        if not isinstance(aVal, (int, float)):
            return
        self._chord = aVal
        self._has_been_set = True
        self.refresh()  


    @property
    def has_been_set (self) -> bool:
        """ True if Reynolds was been set in this dialog """
        return self._has_been_set 


    def _init_layout(self) -> QLayout:

        l = QGridLayout()
        r,c = 0,0 
        SpaceR (l, r, stretch=0, height=5) 
        r += 1
        FieldF  (l,r,c, lab="Chord", width=80, step=10, lim=(10, 9999), dec=0, unit="mm",
                        obj=self, prop=Calc_Reynolds_Dialog.chord,
                        disable=self._chord_fixed)
        Slider  (l,r,c+3, width=80,  lim=(10, 500),   
                        obj=self, prop=Calc_Reynolds_Dialog.chord,
                        hide=self._chord_fixed)
        r += 1
        FieldF  (l,r,c, lab="Velocity", width=80, unit="m/s", step=1, lim=(1, 360), dec=1,
                        obj=self, prop=Calc_Reynolds_Dialog.v)

        Slider  (l,r,c+3, width=80, lim=(1,100), 
                        obj=self, prop=Calc_Reynolds_Dialog.v)
        r += 1
        SpaceR  (l, r, stretch=2, height=5) 
        r += 1
        FieldF  (l,r,c, lab="Reynolds", width=80, unit="k", step=10, lim=(1, 9999), dec=0,
                        obj=self, prop=Calc_Reynolds_Dialog.re_asK)
        r += 1
        Label   (l,r,c+1, style=style.COMMENT, colSpan=5, 
                        get=f"ρ={AIR_RHO}, η={AIR_ETA:.2E} at {TEMP_DEFAULT}°C and {ALT_DEFAULT}m")

        l.setColumnMinimumWidth (0,70)
        l.setColumnMinimumWidth (2,5)
        l.setColumnMinimumWidth (3,50)
        l.setColumnStretch (5,2)   

        return l


    @override
    def _button_box (self):
        """ returns the QButtonBox with the buttons of self"""

        buttons = QDialogButtonBox.StandardButton.Close
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.rejected.connect(self.close)

        return buttonBox 



class Calc_Re_Sqrt_Cl_Dialog (Dialog):
    """ Little dialog to calculate Re.sqrt(Cl) from wing load and chord"""

    _width  = 310
    _height = 200

    name = "Calculate Re·√Cl"

    def __init__ (self, parent : QWidget,  
                  re_asK : int = None,                  # initial Reynolds in 1000
                  fixed_chord : float = None,           # fixed chord length in mm
                  **kwargs): 

        self._has_been_set = False
        self._load   = 40.0                             # wing load in g/dm²
        self._chord  = fixed_chord if fixed_chord is not None else 200
        self._chord_fixed = fixed_chord is not None

        super().__init__ (parent, **kwargs)
        
        if re_asK is not None:
            self.set_re_asK (re_asK)                    # will also set load
            self._has_been_set = False                  # but not yet confirmed by user


    @property
    def re_asK (self) -> int: 
        """ Reynolds number base 1000"""
        return int(re_sqrt_from_load(self.load/10, self.chord/1000, round_to=None) / 1000)
     
    def set_re_asK (self, aVal : float): 

        if not isinstance(aVal, (int, float)):
            return
        self.set_load (load_from_re_sqrt(aVal*1000, self.chord/1000, round_dec=None)*10)   # chord in m


    @property
    def load (self) -> float:
        """ wing load in g/dm² """
        return self._load   
    def set_load (self, aVal : float):
        if not isinstance(aVal, (int, float)):
            return
        self._load = aVal
        self._has_been_set = True
        self.refresh()



    @property
    def chord (self) -> float:
        """ chord in mm """
        return self._chord
    def set_chord (self, aVal : float):
        if not isinstance(aVal, (int, float)):
            return
        self._chord = aVal
        self._has_been_set = True
        self.refresh()  


    @property
    def has_been_set (self) -> bool:
        """ True if Re sqrt(cl) was been set in this dialog """
        return self._has_been_set 


    def _init_layout(self) -> QLayout:

        l = QGridLayout()
        r,c = 0,0 
        SpaceR (l, r, stretch=0, height=5) 
        r += 1
        FieldF  (l,r,c, lab="Chord", width=80, step=10, lim=(10, 9999), dec=0, unit="mm",
                        obj=self, prop=Calc_Re_Sqrt_Cl_Dialog.chord,
                        disable=self._chord_fixed)
        Slider  (l,r,c+3, width=80,  lim=(10, 500),   
                        obj=self, prop=Calc_Re_Sqrt_Cl_Dialog.chord,
                        hide=self._chord_fixed)
        r += 1
        FieldF  (l,r,c, lab="Wing load", width=80, unit="g/dm²", step=1, lim=(1, 999), dec=0,
                        obj=self, prop=Calc_Re_Sqrt_Cl_Dialog.load)

        Slider  (l,r,c+3, width=80, lim=(1,200), 
                        obj=self, prop=Calc_Re_Sqrt_Cl_Dialog.load)
        r += 1
        SpaceR  (l, r, stretch=2, height=5) 
        r += 1
        FieldF  (l,r,c, lab="Re·√Cl", width=80, unit="k", step=10, lim=(1, 999), dec=0,
                        obj=self, prop=Calc_Re_Sqrt_Cl_Dialog.re_asK)
        r += 1
        Label   (l,r,c+1, style=style.COMMENT, colSpan=5, 
                        get=f"ρ={AIR_RHO}, η={AIR_ETA:.2E} at {TEMP_DEFAULT}°C and {ALT_DEFAULT}m")

        l.setColumnMinimumWidth (0,70)
        l.setColumnMinimumWidth (2,5)
        l.setColumnMinimumWidth (3,50)
        l.setColumnStretch (5,2)   

        return l


    @override
    def _button_box (self):
        """ returns the QButtonBox with the buttons of self"""

        buttons = QDialogButtonBox.StandardButton.Close
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.rejected.connect(self.close)

        return buttonBox 




#!/usr/bin/env pythonbutton_color
# -*- coding: utf-8 -*-

"""  

Handle Airfoil UI operations like Open and Select

"""

import os
import fnmatch        

from typing                     import override, Callable    

from PyQt6.QtCore               import QMargins
from PyQt6.QtWidgets            import QHBoxLayout
from PyQt6.QtWidgets            import QFileDialog, QWidget

from ..base.widgets             import * 
from ..base.panels              import MessageBox

from ..model.airfoil            import Airfoil
from ..model.airfoil            import GEO_BASIC
from ..model.airfoil_examples   import Example


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)

# ------ colors of modes ----------------------------------


class mode_color:
    """ colors as kwargs for the different modes """

    MODIFY        = {'color': 'deeppink',          'alpha' : 0.2 }
    OPTIMIZE      = {'color': 'mediumspringgreen', 'alpha' : 0.2 }   


# ----- common methods -----------


def create_airfoil_from_path (parent, pathFilename : str | None,
                              workingDir : str | None = None, 
                              example_if_none=False, 
                              message_delayed=False) -> Airfoil:
    """
    Create and return a new airfoil based on pathFilename.
        Return None if the Airfoil couldn't be loaded 
        or 'Example' airfoil if example_if_none == True  
    """

    airfoil_loaded = False
    airfoil        = None
    msg            = None

    if pathFilename:
        try: 
            airfoil = Airfoil.onFileType(pathFilename, workingDir=workingDir, geometry=GEO_BASIC)
            airfoil.load()
            airfoil_loaded = airfoil.isLoaded
            if not airfoil_loaded:
                msg = f"Airfoil <b>{os.path.basename(pathFilename)}</b> couldn't be loaded."
        except Exception as e:
            msg = f"Airfoil <b>{os.path.basename(pathFilename)}</b> couldn't be loaded: <br>{str(e)}"

    if not airfoil_loaded:

        airfoil = Example() if example_if_none else None

        if pathFilename: 
            
            logger.error (msg)
            using_example = "<br>Using example airfoil." if example_if_none else ""

            if message_delayed:
                QTimer.singleShot (100, lambda: MessageBox.error   (parent,'Load Airfoil', f"{msg}{using_example}", min_height= 60))
            else:
                MessageBox.error   (parent,'Load Airfoil', f"{msg}{using_example}", min_height= 60)

    return airfoil  



def get_airfoil_fileNames_sameDir (airfoil_or_dir : Airfoil | str | None) -> list[str]: 
    """ 
    Returns list of airfoil file names in the same directory as airfoil
        airfoil can be either an Airfoil or a subdirectory
    Returns all .dat, .bez and .hicks files 
    """

    if airfoil_or_dir is None: return []

    if isinstance(airfoil_or_dir, Airfoil):
        airfoil : Airfoil = airfoil_or_dir
        if airfoil.pathFileName_abs is not None: 
            airfoil_dir = os.path.dirname(airfoil.pathFileName_abs) 
        else:
            airfoil_dir = None 
    else: 
        airfoil_dir = airfoil_or_dir

    if not airfoil_dir: airfoil_dir = '.'

    if os.path.isdir (airfoil_dir):
        dat_files = fnmatch.filter(os.listdir(airfoil_dir), '*.dat')
        bez_files = fnmatch.filter(os.listdir(airfoil_dir), '*.bez')
        hh_files  = fnmatch.filter(os.listdir(airfoil_dir), '*.hicks')
        airfoil_files = dat_files + bez_files + hh_files
        return sorted (airfoil_files, key=str.casefold)
    else:
        return []


def get_next_airfoil_in_dir (anAirfoil : Airfoil, example_if_none=False) -> Airfoil: 
    """ 
    Returns next airfoil following anAirfoil in the same directory 
        - or the last airfoil in the directory if anAirfoil was already the last one
        - Example.dat if there are no more airfoil files in the directory 
    """

    airfoil_files = get_airfoil_fileNames_sameDir (anAirfoil)
    next_airfoil  = None 

    # get index 
    try: 
        iair = airfoil_files.index(anAirfoil.fileName)
    except: 
        iair = None

    # get next 
    if iair is not None and len(airfoil_files) > 1: 
        if iair == (len (airfoil_files) - 1):
            next_file = airfoil_files [iair - 1] 
        else:
            next_file = airfoil_files [iair + 1]
    elif iair is not None and len(airfoil_files) == 1: 
        next_file = None
    elif iair is None and len(airfoil_files) > 1:               # find next which is greater than current
        next_file = None
        for air_file in airfoil_files:
            if air_file.lower() > anAirfoil.fileName.lower():
                next_file = air_file
                break
        if next_file is None:   
            next_file = airfoil_files [-1]                      # take last if none greater found
    else: 
        next_file = airfoil_files [0] if airfoil_files else None 

    if next_file: 
        try: 
            next_airfoil = Airfoil.onFileType(next_file, workingDir = anAirfoil.pathName_abs, geometry=GEO_BASIC)
            next_airfoil.load()
        except:
            next_airfoil = None 

    if next_airfoil is None and example_if_none:
        next_airfoil = Example()
 
    return next_airfoil




# ----- widgets -----------


class Airfoil_Select_Open_Widget (Widget, QWidget):
    """ 
    Compound widget to either select or open new airfoil 
        - ComboBox (optional with spin) with files in same directory
        - optional Open button to select new airfoil 
        - optional Delete button to delete current   

    When user successfully selected an airfoil file, 'set' is called with 
    the new <Airfoil> as argument 

    When user deleted, 'set' is called with None 
    """

    _width  = (120, None)

    def __init__(self, *args,
                 get = None,                                # get current / initial airfoil 
                 set : callable= None,                      # callback: will set new airfoil
                 set_open : Callable | None = None,         # alternate callback when opened via button 
                 addEmpty = False,                          # add empty entry to file list 
                 asSpin = False,                            # ComboBox with spin buttons 
                 withOpen = True,                           # include open button 
                 textOpen = "Open",                         # text for Open button 
                 widthOpen = 100,                           # width of open text button 
                 initialDir : Airfoil | str | None = None,  # either an airfoil or a pathString 
                 dialog_parent : QWidget | None = None,     # alternate, explicit parent for QFileDialog if self is transient
                 **kwargs):
        super().__init__(*args, get=get, set=set, **kwargs)

        self._set_open = set_open if callable(set_open) else None

        self._combo_widget  = None 
        self._button_widget = None 
        self._icon_widget   = None 

        self._no_files_here = None
        self._addEmpty = addEmpty is True 
        if isinstance (initialDir, Airfoil):
            airfoil : Airfoil = initialDir
            self._initial_dir = airfoil.pathName_abs
        else:
            self._initial_dir = initialDir

        self._dialog_parent = dialog_parent

        # get initial properties  (cur airfoil) 
        self._get_properties ()

        # build local HBox layout with the guest widget

        l = QHBoxLayout(self)
        l.setContentsMargins (QMargins(0, 0, 0, 0))
        if asSpin: 
            self._combo_widget = ComboSpinBox (l, get=self.airfoil_fileName, 
                                                set=self.set_airfoil_by_fileName, 
                                                options=self.airfoil_fileNames_sameDir,
                                                hide=self.no_files_here,
                                                toolTip=None,               # is set separately 
                                                signal=False)
        else:             
            self._combo_widget = ComboBox      (l, get=self.airfoil_fileName, 
                                                set=self.set_airfoil_by_fileName, 
                                                options=self.airfoil_fileNames_sameDir,
                                                hide=self.no_files_here, 
                                                toolTip=None,               # is set separately 
                                                signal=False)
        if withOpen:

            # either text button if nothing is there
            self._button_widget = Button (l, text=textOpen, set=self._open_airfoil, 
                                                signal=False, width = widthOpen,
                                                hide=lambda: not self.no_files_here() ) 
            # ... or icon button together with combo box 
            self._icon_widget =   ToolButton (l, icon=Icon.OPEN, set=self._open_airfoil, signal=False,
                                                hide=self.no_files_here,
                                                toolTip="Select an airfoil in a different directory")
            l.insertStretch (-1)

        l.setContentsMargins (QMargins(0, 0, 0, 0))

        if self.no_files_here():                                # show only open button
            l.setSpacing (0)
            # l.setStretch (0,0)
            # l.setStretch (2,2)
        else:                                                   # combobox with open 
            l.setSpacing (1)
            l.setStretch (0,2)
            l.setStretch (2,0)

        self.setLayout (l)

        # normal widget handling 

        self._layout_add ()                                     # put into layout - so it gets parent early

        self._get_properties ()
        self._set_Qwidget_static ()
        self._set_Qwidget ()
        self._set_comboBox_tooltip ()


    @override
    def refresh (self, disable=None):

        self._no_files_here = None                              # reset cached value 
        super().refresh(disable) 

        if self._combo_widget is not None:
            self._combo_widget.refresh (disable) 
            self._set_comboBox_tooltip ()
        if self._button_widget is not None:
            self._button_widget.refresh (disable) 
        if self._icon_widget is not None:
            self._icon_widget.refresh (disable) 

        l : QHBoxLayout = self.layout()
        if self.no_files_here():                                # show only open button
            l.setSpacing (0)
            l.setStretch (0,0)
            l.setStretch (2,2)
        else:                                                   # combobox with open 
            l.setSpacing (1)
            l.setStretch (0,2)
            l.setStretch (2,0)


    def _open_airfoil (self):
        """ open a new airfoil and load it"""

        filters   = "Airfoil files (*.dat *.bez);;All files (*.*)"
        directory = self._val.pathName_abs if isinstance (self._val, Airfoil) else self._initial_dir

        # Validate directory, ensure caption - QFileDialog tends to crash with native WIndows Dialog
        if not (directory and os.path.isdir(directory)):
            directory = ""

        # get parent for dialog
        if self._dialog_parent is None:
            parent = self
        else:
            parent = self._dialog_parent            # QFileDialog parent if self is transient (would crash)

        newPathFilename, _ = QFileDialog.getOpenFileName(parent, filter=filters, 
                                                         directory=directory,
                                                         caption="Select Airfoil File to Open")

        if newPathFilename:                         # user pressed open
            airfoil = create_airfoil_from_path (self, newPathFilename)
            if airfoil is not None: 

                if self._set_open is not None:
                    self._set_open (airfoil)                     # call alternate callback
                else:
                    self.set_airfoil (airfoil)



    def no_files_here (self) -> bool:
        """ True if no files to select in combo """

        if self._no_files_here is None: 
            n = len (self.airfoil_fileNames_sameDir())
            if n== 1:
                # there can be a blank or example entry as the first item 
                fileName_first = self.airfoil_fileNames_sameDir()[0]
                no =  fileName_first == ""
            elif n == 0: 
                no =  True
            else: 
                no = False
            self._no_files_here = no 
        return self._no_files_here


    @property
    def airfoil (self) -> Airfoil:
        return self._val
    def set_airfoil (self, anAirfoil : Airfoil):

        self._no_files_here = None                              # reset cached value 
        if isinstance(anAirfoil, Airfoil):                      # new initial for searches
            self._initial_dir = anAirfoil.pathFileName_abs

        self._set_value (anAirfoil)

        # refresh the sub widgets -> hide / show 
        w : Widget
        for w in self.findChildren (Widget):
            w.refresh()

        self._set_comboBox_tooltip()


    def airfoil_fileName (self) -> str | None:
        return self.airfoil.fileName if self.airfoil is not None else None


    def set_airfoil_by_fileName (self, newFileName): 
        """ set new current airfoil bei filename""" 

        # Empty selected - set airfoil to None
        if not newFileName:
            self.set_airfoil (None)
            return 

        # is fileName in directory ?

        fileNames = get_airfoil_fileNames_sameDir (self.airfoil if self.airfoil else self._initial_dir)

        if newFileName in fileNames: 

            # create and set new airfoil

            workingDir = self.airfoil.pathName_abs if self.airfoil else self._initial_dir

            new_airfoil = create_airfoil_from_path (self, newFileName, workingDir=workingDir)
 
            if new_airfoil:
                self.set_airfoil (new_airfoil)
        

    def airfoil_fileNames_sameDir (self): 
        """ list of airfoil filenames in the same dir as current airfoil"""


        fileNames = get_airfoil_fileNames_sameDir (self.airfoil if self.airfoil else self._initial_dir)

        if self._addEmpty: 
            fileNames.insert (0,"")  
        return fileNames


    def _set_comboBox_tooltip (self):
        """ set tooltip for comboBox - normally pathFileName of airfoil"""

        if self._combo_widget:

            toolTip = self.airfoil.info_as_html if self.airfoil else self._toolTip 

            self._combo_widget.setToolTip (toolTip)  



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

UI panels 

"""

import logging

from PyQt6.QtWidgets        import QMenu
from PyQt6.QtGui            import QDesktopServices
from PyQt6.QtCore           import QUrl

from ..base.widgets         import * 
from ..base.panels          import Edit_Panel

from ..model.airfoil        import Airfoil
from ..model.airfoil_geometry import (Geometry, Geometry_Bezier, Curvature_Abstract,
                                    Line, Side_Airfoil_Bezier)
from ..model.case           import Case_Abstract, Case_Direct_Design
from ..model.xo2_driver     import Xoptfoil2

from .ae_widgets            import * 
from .ae_dialogs            import (Match_Bezier_Dialog, Matcher, LE_Radius_Dialog, TE_Gap_Dialog,
                                   Blend_Airfoil_Dialog, Flap_Airfoil_Dialog, Repanel_Airfoil_Dialog)

from ..app_model            import App_Model

logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


class Panel_Airfoil_Abstract (Edit_Panel):
    """ 
    Abstract superclass for Edit/View-Panels of AirfoilEditor
        - has semantics of App
        - connect / handle signals 
    """

    sig_toggle_panel_size = pyqtSignal()                # wants to toggle panel size

    MAIN_MARGINS        = (10, 5,20, 5)
    MAIN_MARGINS_MINI   = ( 0, 5,10, 5)
    MAIN_MARGINS_FILE   = (10, 5,10, 5)

    _main_margins = MAIN_MARGINS

    @property
    def app_model (self) -> App_Model:
        return self.dataObject

    @property
    def airfoil (self) -> Airfoil: 
        return self.app_model.airfoil

    @property
    def geo (self) -> Geometry:
        return self.airfoil.geo

    @property    
    def case (self) -> Case_Abstract:
        return self.app_model.case


    @property
    def is_mode_modify (self) -> bool:
        """ panel in mode_modify or disabled ? """ 
        return self.app_model.is_mode_modify or self.app_model.is_mode_as_bezier


    @property
    def is_mode_optimize (self) -> bool:
        """ panel in mode_optimize or disabled ? """
        return self.app_model.is_mode_optimize


    @property
    def is_bezier (self) -> bool:
        """ True if self is in mode_modify and geo is Bezier """
        return self.airfoil.isBezierBased if self.airfoil else False


    @override
    @property
    def _isDisabled (self) -> bool:
        """ overloaded: only enabled in edit mode of App """
        return not self.is_mode_modify or (self.airfoil.isFlapped if self.airfoil else False)
    
    @override
    def _set_panel_layout (self ):
        """ Set layout of self._panel """
        # overridden to connect to widgets changed signal

        super()._set_panel_layout ()
        for w in self.widgets:
            w.sig_changed.connect (self._on_widget_changed)
        for w in self.header_widgets:
            w.sig_changed.connect (self._on_widget_changed)


    def _on_widget_changed (self, widget):
        """ user changed data in widget"""
        logger.debug (f"{self} {widget} widget changed slot")
        self.app_model.notify_airfoil_changed ()


# --------------------------------------------------------------------------


class Panel_File_View (Panel_Airfoil_Abstract):
    """ File panel with open / save / ... """

    name = 'View Mode'

    sig_modify = pyqtSignal()                           # wants to modify the airfoil
    sig_optimize = pyqtSignal()                         # wants to enter optimize mode
    sig_exit = pyqtSignal()                             # wants to exit the application
    sig_new_as_bezier = pyqtSignal()                    # wants to create new Bezier based airfoil
    sig_save_as = pyqtSignal()                          # wants to save current airfoil as new file
    sig_rename = pyqtSignal()                           # wants to rename current airfoil
    sig_delete = pyqtSignal()                           # wants to delete current airfoil
    sig_delete_temp_files = pyqtSignal()                # wants to delete all temp files
    sig_load_airfoil_settings = pyqtSignal()            # load airfoil settings if there are any

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_FILE
    
    @override
    @property
    def _isDisabled (self) -> bool:
        """ override: enabled (as parent data panel is disabled)"""
        return False
    

    def _on_widget_changed (self, *_ ):
        """ user changed data in widget"""
        # overloaded - do not react on self widget changes 
        pass


    @override
    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""
        l_head.addStretch(1)
        ToolButton   (l_head, icon=Icon.EXPAND, set=self.sig_toggle_panel_size.emit,
                      toolTip='Minimize lower panel -<br>Alternatively, you can double click on the lower panels')


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        w =Airfoil_Select_Open_Widget (l,r,c, colSpan=4, signal=False, 
                                    textOpen="&Open", widthOpen=100, 
                                    get=lambda: self.airfoil, 
                                    set=     lambda airfoil: self.app_model.set_airfoil(airfoil),
                                    set_open=lambda airfoil: self.app_model.set_airfoil(airfoil, load_settings=True))

        r += 1
        Button (l,r,c, text="&Modify", width=100, 
                set=self.sig_modify.emit, toolTip="Modify geometry, Normalize, Repanel, Set Flap",
                button_style=button_style.PRIMARY)
        MenuButton (l,r,c+2, text="More...", width=80, 
                menu=self._more_menu(), 
                toolTip="Choose further actions for this airfoil")
        r += 1
        Button (l,r,c, text="&Optimize...", width=100, 
                set=self.sig_optimize.emit, toolTip=self._tooltip_optimize,
                disable=lambda: not Xoptfoil2.ready)
        r += 1
        l.setRowStretch (r,2)
        r += 1
        Button (l,r,c, text="&Exit", width=100, set=self.sig_exit.emit)
        l.setColumnStretch (2,2)
        l.setColumnMinimumWidth (1,12)

        return l 
 

    def _more_menu (self) -> QMenu:
        """ create and return sub menu for 'more' actions"""

        menu = QMenu ()

        menu.addAction (MenuAction ("As Bezier based", self, set=self.sig_new_as_bezier, 
                                     disable=lambda: self.airfoil.isBezierBased,
                                     toolTip="Create new Bezier based airfoil of current airfoil"))
        menu.addSeparator ()
        menu.addAction (MenuAction ("Save as...", self, set=self.sig_save_as.emit,
                                     toolTip="Create a copy of the current airfoil with new name and filename"))
        menu.addAction (MenuAction ("Rename...", self, set=self.sig_rename.emit,
                                     toolTip="Rename name and/or filename of current airfoil"))
        menu.addAction (MenuAction ("Delete", self, set=self.sig_delete.emit,
                                     toolTip="Delete current airfoil including all temporary files created by the AirfoilEditor"))
        menu.addAction (MenuAction ("Delete temp files", self, set=self.sig_delete_temp_files.emit,
                                     toolTip="Delete all temporary files created by the AirfoilEditor just to have a clean directory again"))
        menu.addSeparator ()
        menu.addAction (MenuAction ("Readme on Github", self, set=self._open_AE_url,
                                     toolTip="Open the Github README file of the AirfoilEditor in a browser"))
        menu.addAction (MenuAction ("Releases on Github", self, set= self._open_releases_url,
                                     toolTip="Open the Github page with the actual release of the AirfoilEditor"))

        menu.setToolTipsVisible(True)

        return menu


    def _open_releases_url (self):
        """ open Github versions in Browser"""

        link = "https://github.com/jxjo/AirfoilEditor/releases"
        QDesktopServices.openUrl(QUrl(link))


    def _open_AE_url (self):
        """ open Github AirfoilEditor repo in Browser"""

        link = "https://github.com/jxjo/AirfoilEditor"
        QDesktopServices.openUrl(QUrl(link))


    def _tooltip_optimize (self) -> str:
        """ dynamic tooltip to handle xo2 not ready"""

        if Xoptfoil2.ready:
            return f"Optimize an airfoil with {Xoptfoil2.NAME}"
        else: 
            return f"Optimization not possible <br><br><b>{Xoptfoil2.NAME} not ready</b><br><br>{Xoptfoil2.ready_msg}"



class Panel_File_View_Small (Panel_File_View):
    """ View File panel in small version """

    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        Airfoil_Select_Open_Widget (l,r,c, colSpan=4, signal=False, 
                                    textOpen="&Open", widthOpen=100, 
                                    get=lambda: self.airfoil, 
                                    set=     lambda airfoil: self.app_model.set_airfoil(airfoil),
                                    set_open=lambda airfoil: self.app_model.set_airfoil(airfoil, load_settings=True))
        r += 1
        Button      (l,r,c, text="&Exit", width=100, set=self.sig_exit.emit)
        MenuButton  (l,r,c+2, text="More...", width=80, 
                        menu=self._more_menu(), 
                        toolTip="Choose further actions for this airfoil")
        ToolButton  (l,r,c+3, icon=Icon.COLLAPSE, set=self.sig_toggle_panel_size.emit,
                        toolTip='Maximize lower panel -<br>Alternatively, you can double click on the lower panels')

        l.setColumnMinimumWidth (1,12)
        l.setColumnStretch (2,2)
        return l 
 

    def _more_menu (self) -> QMenu:
        """ create and return sub menu for 'more' actions"""

        menu = super()._more_menu ()
        menu.insertAction (menu.actions()[0], MenuAction ("&Optimize", self, set=self.sig_optimize.emit,   
                                    toolTip="Modify geometry, Normalize, Repanel, Set Flap"))
        menu.insertAction (menu.actions()[0], MenuAction ("&Modify", self, set=self.sig_modify.emit,   
                                    toolTip=self._tooltip_optimize(), disable=lambda: not Xoptfoil2.ready))
        return menu



class Panel_File_Modify (Panel_Airfoil_Abstract):
    """ File panel with open / save / ... """

    name = 'Modify Mode'

    sig_finish   = pyqtSignal()                                 # wants to finish modify mode - ok / cancel
    sig_cancel   = pyqtSignal()                                 # wants to cancel modify mode

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_FILE


    @override
    def title_text (self) -> str: 
        """ returns text of title - default self.name"""
        if self.airfoil and self.airfoil.isBezierBased:
             return 'Bezier Mode'
        else: 
            return self.name

    @override
    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""
        l_head.addStretch(1)
        ToolButton   (l_head, icon=Icon.EXPAND, set=self.sig_toggle_panel_size.emit,
                      toolTip='Minimize lower panel -<br>Alternatively, you can double click on the lower panels')


    @property
    def _isDisabled (self) -> bool:
        """ override: always enabled """
        return False
    
    @property
    def case (self) -> Case_Direct_Design:
        return super().case


    def _init_layout (self): 

        self.set_background_color (**mode_color.MODIFY)

        l = QGridLayout()
        r,c = 0, 0 
        Field (l,r,c, colSpan=5,  get=lambda: self.case.airfoil_seed.fileName,
                        toolTip="File name of seed airfoil for design airfoils")
        r += 1
        ComboSpinBox (l,r,c, colSpan=3, width=146, get=self.airfoil_fileName, 
                        set=self.set_airfoil_by_fileName,
                        options=self.airfoil_fileNames, signal=False,
                        toolTip="Select current design airfoil to modify")
        ToolButton   (l,r,c+3, icon=Icon.DELETE, set=self.app_model.remove_current_design,
                        hide=lambda: self.case.get_iDesign (self.airfoil) == 0,  # hide Design 0 
                        toolTip="Remove current design airfoil")
        r += 1
        l.setRowStretch (r,2)
        r += 1
        Button (l,r,c,  text="&Finish ...", width=100, set=self.sig_finish.emit, 
                        toolTip="Save current airfoil, optionally modify name and leave edit mode")
        r += 1
        Button (l,r,c,  text="&Cancel",  width=100, 
                        set=self.sig_cancel.emit,
                        toolTip="Cancel modifications of airfoil and leave edit mode")
        l.setColumnMinimumWidth (1,12)
        l.setColumnStretch (4,2)

        return l


    def airfoil_fileName(self) -> list[str]:
        """ fileName of current airfoil without extension"""
        return os.path.splitext(self.airfoil.fileName)[0]


    def airfoil_fileNames(self) -> list[str]:
        """ list of design airfoil fileNames without extension"""

        fileNames = []
        if self.case:
            for airfoil in self.case.airfoil_designs:
                fileNames.append (os.path.splitext(airfoil.fileName)[0])
        return fileNames


    def set_airfoil_by_fileName (self, fileName : str):
        """ set new current design airfoil by fileName"""

        airfoil = self.case.get_design_by_name (fileName)
        self.app_model.set_airfoil (airfoil)




class Panel_File_Modify_Small (Panel_File_Modify):
    """ Modify File panel in small version """

    def _init_layout (self): 

        self.set_background_color (**mode_color.MODIFY)

        l = QGridLayout()
        r,c = 0, 0 
        ComboSpinBox (l,r,c, colSpan=3, width=146, get=self.airfoil_fileName, 
                        set=self.set_airfoil_by_fileName,
                        options=self.airfoil_fileNames, signal=False,
                        toolTip="Select current design airfoil to modify")
        ToolButton   (l,r,c+3, icon=Icon.DELETE, set=self.app_model.remove_current_design,
                        hide=lambda: self.case.get_iDesign (self.airfoil) == 0,
                        toolTip="Remove current design airfoil")  
        ToolButton  (l,r,c+5, icon=Icon.COLLAPSE, set=self.sig_toggle_panel_size.emit,
                        toolTip='Maximize lower panel -<br>Alternatively, you can double click on the lower panels')
        r += 1
        Button (l,r,c,  text="&Finish ...", width=100, set=self.sig_finish.emit, 
                        toolTip="Save current airfoil, optionally modify name and leave edit mode")
        Button (l,r,c+2,text="&Cancel",  width=80, colSpan=3,
                        set=lambda : self.sig_cancel.emit(),
                        toolTip="Cancel modifications of airfoil and leave edit mode")
        l.setColumnMinimumWidth (1,12)
        l.setColumnStretch (4,2)
        return l



class Panel_Geometry (Panel_Airfoil_Abstract):
    """ Main geometry data of airfoil"""

    name = 'Geometry'

    @override
    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        # blend with airfoil - currently Bezier is not supported
        Button (l_head, text="&Blend", width=75,
                set=self.do_blend_with, 
                hide=lambda: not self.is_mode_modify or self.airfoil.isBezierBased,
                toolTip="Blend original airfoil with another airfoil")


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        FieldF (l,r,c, lab="Thickness", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_thick,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="Camber", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_camb,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        r += 1
        FieldF (l,r,c, lab="LE radius", width=75, unit="%", step=0.02,
                obj=lambda: self.geo, prop=Geometry.le_radius, disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self.do_le_radius, 
                hide=lambda: not self.is_mode_modify or self.is_bezier,
                toolTip="Set leading edge radius with a flexible blending range")
        r += 1
        FieldF (l,r,c, lab="TE gap", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.te_gap, disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self.do_te_gap,
                hide=lambda: not self.is_mode_modify or self.is_bezier,
                toolTip="Set trailing edge gap with a flexible blending range")
        r += 1
        SpaceR (l,r, height=5)
        r += 1
        Label  (l,r,0,colSpan=5, get=self._messageText, style=style.COMMENT, height=(None,None))
        l.setColumnMinimumWidth (c,80)
        l.setColumnMinimumWidth (c+2,30)
        l.setColumnStretch (c+3,2)

        r,c = 0,4 
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_thick_x,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_camb_x,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        r += 1
        FieldF (l,r,c, lab="LE curv", width=75, dec=0, disable=True,
                obj=lambda: self.geo.curvature, prop=Curvature_Abstract.max)
        r += 1
        FieldF (l,r,c, lab="TE curv", width=75, dec=0, disable=True,
                obj=lambda: self.geo.curvature, prop=Curvature_Abstract.max_te,
                style=lambda: style.NORMAL if self.geo.curvature.max_te <2 else style.WARNING)

        l.setColumnMinimumWidth (4,60)
        return l 


    def _messageText (self): 
        """ text to show at bottom of panel"""

        if self.geo.curvature.max_te >= 2:
            text = f"- Curvature at trailing edge is quite high"
        else:
            text = f"Geometry {self.geo.description}"
        return text 


    def do_blend_with (self): 
        """ blend with another airfoil - open blend airfoil dialog """ 

        dialog = Blend_Airfoil_Dialog (self, self.airfoil, self.app_model.airfoil_seed, 
                                       parentPos=(0.75, 0.2), dialogPos=(0,1))  

        dialog.sig_airfoil_2_changed.connect    (self.app_model.set_airfoil_2)
        dialog.sig_blend_changed.connect        (self.app_model.notify_airfoil_geo_changed)

        dialog.exec()     

        if dialog.airfoil2 is not None: 
            # do final blend with high quality (splined) 
            self.airfoil.geo.blend (self.app_model.airfoil_seed.geo, 
                                      dialog.airfoil2.geo, 
                                      dialog.blendBy) 
            self.app_model.set_airfoil_2 (None)
            self.app_model.notify_airfoil_changed()



    def do_le_radius (self): 
        """ set LE radius - run set LE radius dialog""" 

        if self.airfoil.isBezierBased: return                   # not for Bezier airfoils

        dialog = LE_Radius_Dialog (self, self.app_model, parentPos=(0.25, 0.75), dialogPos=(0,1))
        dialog.exec()     


    def do_te_gap (self): 
        """ set TE gap - run set TE gap dialog""" 

        if self.airfoil.isBezierBased: return                   # not for Bezier airfoils

        dialog = TE_Gap_Dialog (self, self.app_model, parentPos=(0.25, 0.75), dialogPos=(0,1))
        dialog.exec()     



class Panel_Geometry_Small (Panel_Geometry):
    """ Main geometry data of airfoil - small version"""

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        FieldF (l,r,c, lab="Thickness", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_thick,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="Camber", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_camb,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        l.setColumnMinimumWidth (c,70)
        l.setColumnMinimumWidth (c+2,10)
        r,c = 0, 3 
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_thick_x,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_camb_x,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        l.setColumnMinimumWidth (c,20)
        c += 2
        l.setColumnMinimumWidth (c,15)
        c += 1
        r = 0
        FieldF (l,r,c, lab="LE radius", width=60, unit="%", step=0.02,
                obj=lambda: self.geo, prop=Geometry.le_radius, disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self.do_le_radius, 
                hide=lambda: not self.is_mode_modify or self.is_bezier,
                toolTip="Set leading edge radius with a flexible blending range")
        r += 1
        FieldF (l,r,c, lab="TE gap", width=60, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.te_gap, disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self.do_te_gap,
                hide=lambda: not self.is_mode_modify or self.is_bezier,
                toolTip="Set trailing edge gap with a flexible blending range")
        l.setColumnMinimumWidth (c,70)
        l.setColumnStretch (c+3,2)
        return l 



class Panel_Panels (Panel_Airfoil_Abstract):
    """ Panelling information """

    name = 'Panels'

    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        # repanel airfoil - currently Bezier is not supported
        Button (l_head, text="&Repanel", width=75,
                set=self.do_repanel, 
                hide=lambda: not self.is_mode_modify,
                disable=lambda: self.geo.isBasic or self.geo.isHicksHenne,
                toolTip="Repanel airfoil with a new number of panels" ) 


    def _init_layout (self):

        l = QGridLayout()
        r,c = 0, 0 
        FieldI (l,r,c, lab="No of panels", disable=True, width=60, style=self._style_panel,
                get=lambda: self.geo.nPanels, )
        r += 1
        FieldF (l,r,c, lab="Angle at LE", width=60, dec=1, unit="°", style=self._style_angle,
                get=lambda: self.geo.panelAngle_le)
        SpaceC (l,c+2, width=10, stretch=0)
        Label  (l,r,c+3,width=70, get=lambda: f"at index {self.geo.iLe}", style=style.COMMENT)
        r += 1
        FieldF (l,r,c, lab="Angle min", width=60, dec=1, unit="°",
                get=lambda: self.geo.panelAngle_min[0])
        Label  (l,r,c+3, get=lambda: f"at index {self.geo.panelAngle_min[1]}", style=style.COMMENT)
        r += 1
        SpaceR (l,r,height=5)
        r += 1
        Label  (l,r,0,colSpan=4, get=self._messageText, style=style.COMMENT, height=(None,None))

        l.setColumnMinimumWidth (0,80)
        l.setColumnStretch (c+4,1)    
        return l
 

    def do_repanel (self): 
        """ repanel airfoil - open repanel dialog""" 

        dialog = Repanel_Airfoil_Dialog (self, self.app_model,
                                         parentPos=(0.35, 0.75), dialogPos=(0,1))
        dialog.exec()     


    def _on_panelling_finished (self, aSide : Side_Airfoil_Bezier):
        """ slot for panelling (dialog) finished - reset airfoil"""


    def _style_panel (self):
        """ returns style.WARNING if panels not in range"""
        if self.geo.nPanels < 120 or self.geo.nPanels > 260: 
            return style.WARNING
        else: 
            return style.NORMAL


    def _style_angle (self):
        """ returns style.WARNING if panel angle too blunt"""
        if self.geo.panelAngle_le > Geometry.LE_PANEL_ANGLE_TOO_BLUNT: 
            return style.WARNING
        elif self.geo.panelAngle_le < Geometry.PANEL_ANGLE_TOO_SHARP: 
            return style.WARNING
        else: 
            return style.NORMAL

    def _messageText (self): 

        text = []
        minAngle, iAngle = self.geo.panelAngle_min

        if self.geo.panelAngle_le == 180.0: 
            text.append("- Leading edge has 2 points")
        elif self.geo.panelAngle_le > Geometry.LE_PANEL_ANGLE_TOO_BLUNT: 
            text.append(f"- Panel angle at LE {self.geo.panelAngle_le:.1f}° is too blunt")

        if self.geo.panelAngle_le < Geometry.PANEL_ANGLE_TOO_SHARP: 
            text.append(f"- Panel angle at LE {self.geo.panelAngle_le:.1f}° is too sharp")
        elif minAngle < Geometry.PANEL_ANGLE_TOO_SHARP: 
            text.append(f"- Panel angle at i={iAngle} is < {Geometry.PANEL_ANGLE_TOO_SHARP}°")

        if self.geo.nPanels < 100 or self.geo.nPanels > 200: 
            text.append("- No of panels should be > 100 and < 200")
        
        text = '\n'.join(text)
        return text 



class Panel_Panels_Small (Panel_Panels):
    """ Panelling information - small version"""

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self):

        l = QGridLayout()
        r,c = 0, 0 
        FieldI (l,r,c, lab="No of panels", disable=True, width=60, style=self._style_panel,
                get=lambda: self.geo.nPanels, )
        ToolButton (l,r,c+2, icon=Icon.EDIT,
                set=self.do_repanel, hide=lambda: not self.is_mode_modify,
                disable=lambda: self.geo.isBasic or self.geo.isHicksHenne,
                toolTip="Repanel airfoil with a new number of panels" ) 
        r += 1
        l.setRowStretch (r,2)
        l.setColumnMinimumWidth (0,80)
        return l
 


class Panel_Flap (Panel_Airfoil_Abstract):
    """ Flap information and set flap"""

    name = 'Flap'
    _width  = (160,250)

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        isProbablyFlapped = self.airfoil.geo.isProbablyFlapped if self.airfoil else False
        return (self.is_mode_modify or isProbablyFlapped) and not self.is_bezier


    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        Button (l_head, text="Set F&lap", width=75,
                set=self.do_flap, hide=lambda: not self.is_mode_modify,
                disable=self._set_flap_disabled,
                toolTip="Set flap at airfoil" ) 


    def _set_flap_disabled (self) -> bool:
        """ True if set flap is not possible"""

        if self.geo.isBezier or self.geo.isHicksHenne: 
            return True
        elif self.is_mode_modify:
            return not self.airfoil.flap_setter                       # no flapper, no set flap 
        else:
            return True


    def _init_layout (self):

        geo             = self.airfoil.geo

        l = QGridLayout()
        r,c = 0, 0 

        if self.is_mode_modify:

            flap_setter         = self.airfoil.flap_setter
            airfoil_flapped = flap_setter.airfoil_flapped if flap_setter else None

            if airfoil_flapped:
                FieldF (l,r,c, lab="Hinge x", width=60, get=lambda: flap_setter.x_flap, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: flap_setter.flap_angle)
                r +=1
                Field (l,r,c, lab="Based on", colSpan=2,
                       get=flap_setter.airfoil_base.fileName)
                r += 1
                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, height=(40,None), colSpan=3, style=style.COMMENT, wordWrap=True,
                        get="As the base airfoil is available, another flap setting can be applied.")
                lab.setAlignment (ALIGN_BOTTOM)
                l.setRowStretch (r,1)
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)

            elif flap_setter:

                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None,  colSpan=3, style=style.COMMENT, wordWrap=True,
                             get="No flap set")
                l.setColumnStretch (2,3)

            elif self.airfoil.isFlapped:
                FieldF (l,r,c, lab="Hinge x", width=60, get=lambda: geo.curvature.flap_kink_at, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r +=1
                SpaceR (l,r, stretch=2)
                r += 1
                Label  (l,r,c, width=None, colSpan=3, style=style.WARNING,
                             styleRole=QPalette.ColorRole.Window, 
                             get="The airfoil is already flapped.")
                r += 1
                lab =Label  (l,r,c, width=None, height=(40,None), colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="Choose a Design, which is not flapped to set a flap.")
                lab.setAlignment (ALIGN_BOTTOM)
                l.setRowStretch (r,1)
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)

        else: 
            # mode not modify - info about a flap which is (could be) set
            
            if geo.isFlapped:

                FieldF (l,r,c, lab="Hinge x", width=60, get=lambda: geo.curvature.flap_kink_at, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r +=1
                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="The airfoil has a set flap.")
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)
                                
            elif geo.isProbablyFlapped:

                SpaceR (l,r, stretch=2)
                Label  (l,r,c, width=160, height=(None,None), colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="The airfoil is probably flapped, but a kink in the contour couldn't be detected on both sides.")
                r += 1
                l.setRowStretch (r,1)

        return l


    def do_flap (self): 
        """ set flaps - run set flap dialog""" 

        dialog = Flap_Airfoil_Dialog (self, self.app_model, parentPos=(0.55, 0.80), dialogPos=(0,1.2))
        dialog.exec()     


    @override
    def refresh(self, reinit_layout=False):

        # force new layout to show different flap states 
        return super().refresh(reinit_layout=True)


class Panel_Flap_Small (Panel_Flap):
    """ Flap information and set flap"""

    _width  = None
    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self):

        geo             = self.airfoil.geo

        l = QGridLayout()
        r,c = 0, 0 

        if self.is_mode_modify:

            flap_setter         = self.airfoil.flap_setter
            airfoil_flapped = flap_setter.airfoil_flapped if flap_setter else None

            if airfoil_flapped:
                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: flap_setter.flap_angle)

            elif flap_setter:
                Label  (l,r,c, width=None,  colSpan=3, style=style.COMMENT, wordWrap=True,
                             get="No flap set")

            elif self.airfoil.isFlapped:
                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r += 1
                Label  (l,r,c, width=None, colSpan=3, style=style.WARNING,
                             styleRole=QPalette.ColorRole.Window, 
                             get="The airfoil is already flapped.")
 
            ToolButton (l,0,c+2, icon=Icon.EDIT,
                    set=self.do_flap, hide=lambda: not self.is_mode_modify,
                    disable=self._set_flap_disabled,
                    toolTip="Set flap at airfoil" ) 
            l.setRowStretch (r+1,1)
            l.setColumnMinimumWidth (0,80)

        else: 
            # mode not modify - info about a flap which is (could be) set
            
            if geo.isFlapped:

                FieldF (l,r,c, lab="Flap Angle", width=60, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r += 1
                FieldF (l,r,c, lab="Hinge x", width=60, get=lambda: geo.curvature.flap_kink_at, dec=1, unit="%")
                l.setColumnMinimumWidth (0,80)
                                
            elif geo.isProbablyFlapped:

                Label  (l,r,c, width=None, colSpan=2, style=style.COMMENT, 
                             get="The airfoil is probably flapped")
                l.setRowStretch (1,1)

        return l



class Panel_LE_TE  (Panel_Airfoil_Abstract):
    """ info about LE and TE coordinates"""

    name = 'LE, TE'

    _width  = None

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is not Bezier """
        return not self.is_bezier 


    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        Button (l_head, text="&Normalize", width=75,
                set=self.do_normalize, 
                hide=lambda: not self.is_mode_modify,
                toolTip="Normalize airfoil to get leading edge at 0,0 and trailing edge at x=1.0")


    def _init_layout (self): 

        l = QGridLayout()     
        r,c = 0, 0 
        FieldF (l,r,c, lab="Leading edge x,y", get=lambda: self.geo.le[0], width=75, dec=7, style=lambda: self._style (self.geo.le[0], 0.0))
        r += 1
        FieldF (l,r,c, lab="  ... of spline", get=lambda: self.geo.le_real[0], width=75, dec=7, style=self._style_le_real,
                hide=lambda: not self.is_mode_modify)
        r += 1
        FieldF (l,r,c, lab="Trailing edge x,y", get=lambda: self.geo.te[0], width=75, dec=7, style=lambda: self._style (self.geo.te[0], 1.0))
        r += 1
        FieldF (l,r,c,lab="  ... lower", get=lambda: self.geo.te[2], width=75, dec=7, style=lambda: self._style (self.geo.te[0], 1.0))
        l.setColumnMinimumWidth (0,95)
        l.setColumnMinimumWidth (2,10)
        l.setColumnStretch (2,1)
        r,c = 0, 3 
        FieldF (l,r,c+1,get=lambda: self.geo.le[1], width=75, dec=7, style=lambda: self._style (self.geo.le[1], 0.0))
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.le_real[1], width=75, dec=7, style=self._style_le_real,
                hide=lambda: not self.is_mode_modify)
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.te[1], width=75, dec=7, style=lambda: self._style (self.geo.te[1], -self.geo.te[3]))
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.te[3], width=75, dec=7, style=lambda: self._style (self.geo.te[3], -self.geo.te[1]))
        r += 1
        SpaceR (l,r, height=5)
        r += 1
        Label  (l,r,0,colSpan=5, get=self._messageText, style=style.COMMENT, height=(None,None))
        return l


    def _style_le_real (self):
        """ returns style.WARNING if LE spline isn't close to LE"""
        if self.geo.isLe_closeTo_le_real: 
            if self.geo.isBasic:
                return style.NORMAL
            else: 
                return style.NORMAL
        else: 
            return style.WARNING


    def _style (self, val, target_val):
        """ returns style.WARNING if val isn't target_val"""
        if val != target_val and not self.airfoil.isFlapped: 
            return style.WARNING
        else: 
            return style.NORMAL


    def _messageText (self): 

        text = []
        te_not_at_1 = ""
        te_not_sym  = ""
        if not self.geo.isNormalized:
            if self.geo.isSplined and not self.geo.isLe_closeTo_le_real:
                text.append("- LE of spline is not at 0,0")
            elif self.geo.le[0] != 0.0 or self.geo.le[1] != 1.0 : 
                text.append("- LE is not at 0,0")
        if not self.airfoil.isFlapped:
            if self.geo.te[0] != 1.0 or self.geo.te[2] != 1.0 : 
                te_not_at_1 = "- TE x is not at 1.0"
            if self.geo.te[1] != -self.geo.te[3]: 
                if te_not_at_1:
                    te_not_at_1 += " and y is not symmetrical"
                else:   
                    te_not_sym = "- TE y is not symmetrical"
            if te_not_at_1:
                text.append (te_not_at_1)
            if te_not_sym:
                text.append (te_not_sym)

        if not text:
            if self.geo.isSymmetrical: 
                text.append("Airfoil is symmetrical")
            else: 
                text.append("Airfoil is normalized")

        text = '\n'.join(text)
        return text 


    def do_normalize (self):
        """ normalize airfoil to LE at 0,0 and TE at x=1.0"""

        self.airfoil.normalize()
        self.app_model.notify_airfoil_changed()



class Panel_LE_TE_Small  (Panel_LE_TE):
    """ info about LE and TE coordinates - small version"""

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self): 

        l = QGridLayout()     
        r,c = 0, 0 
        FieldF (l,r,c, lab="LE x,y", get=lambda: self.geo.le[0], width=75, dec=7, style=lambda: self._style (self.geo.le[0], 0.0))
        r += 1
        FieldF (l,r,c, lab="TE xm,ym",  width=75, dec=7, 
                get=lambda: (self.geo.te[0] + self.geo.te[2]) / 2, style=lambda: self._style (self.geo.te[0], 1.0))
        l.setColumnMinimumWidth (0,70)
        l.setColumnMinimumWidth (2,10)
        r,c = 0, 3 
        FieldF (l,r,c+1,get=lambda: self.geo.le[1], width=75, dec=7, style=lambda: self._style (self.geo.le[1], 0.0))
        r += 1
        FieldF (l,r,c+1,get=lambda: (self.geo.te[1] + self.geo.te[3])/2, width=75, dec=7, style=lambda: self._style (self.geo.te[1], -self.geo.te[3]))

        r,c = 0, 5 
        ToolButton (l,r,c+1, icon=Icon.EDIT,
                set=self.do_normalize, 
                hide=lambda: not self.is_mode_modify,
                toolTip="Normalize airfoil to get leading edge at 0,0 and trailing edge at x=1.0")
        return l



class Panel_Bezier (Panel_Airfoil_Abstract):
    """ Info about Bezier curves upper and lower  """

    name = 'Bezier'

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        return self.is_bezier

    @override
    @property
    def geo (self) -> Geometry_Bezier:
        return super().geo

    @property
    def upper (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier:
            return self.geo.upper

    @property
    def lower (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier:
            return self.geo.lower


    def _init_layout (self):

        l = QGridLayout()
        r,c = 0, 0 
        Label (l,r,c, get="Bezier control Points", colSpan=4)
        r += 1
        FieldI (l,r,c,   lab="Upper side", get=lambda: self.upper.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.upper, n))
        r += 1
        FieldI (l,r,c,   lab="Lower side",  get=lambda: self.lower.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.lower, n))
        r += 1
        l.setRowStretch (r,2)
        l.setColumnMinimumWidth (0,80)
        return l
 

class Panel_Bezier_Small (Panel_Bezier):
    """ Info about Bezier curves upper and lower - small version """

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self):

        l = QGridLayout()
        r,c = 0, 0 
        FieldI (l,r,c,   lab="Bezier Upper", get=lambda: self.upper.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.upper, n))
        r += 1
        FieldI (l,r,c,   lab="Bezier Lower",  get=lambda: self.lower.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.lower, n))
        l.setColumnMinimumWidth (0,80)
        return l



class Panel_Bezier_Match (Panel_Airfoil_Abstract):
    """ Match Bezier functions  """

    name = 'Bezier Match'

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        return self.is_bezier

    @property
    def upper (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier: return self.geo.upper

    @property
    def lower (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier: return self.geo.lower

    @property
    def curv_upper (self) -> Line:
        if self.geo.isBezier:
            return self.geo.curvature.upper

    @property
    def curv_lower (self) -> Line:
        if self.geo.isBezier:
            return self.geo.curvature.lower

    @property
    def curv (self) -> Curvature_Abstract:
        return self.geo.curvature

    @property
    def target_airfoil (self) -> Airfoil:
        return self.app_model.airfoil_seed

    @property
    def target_upper (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.upper

    @property
    def target_lower (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.lower

    @property
    def target_curv_le (self) -> float:
        return self.target_airfoil.geo.curvature.best_around_le

    @property
    def max_curv_te_upper (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.curvature.at_upper_te

    @property
    def max_curv_te_lower (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.curvature.at_lower_te

    def norm2_upper (self): 
        """ norm2 deviation of airfoil to target - upper side """
        if self._norm2_upper is None: 
            self._norm2_upper = Line.norm2_deviation_to (self.upper.bezier, self.target_upper) 
        return  self._norm2_upper    


    def norm2_lower (self): 
        """ norm2 deviation of airfoil to target  - upper side """
        if self._norm2_lower is None: 
            self._norm2_lower = Line.norm2_deviation_to (self.lower.bezier, self.target_lower)  
        return self._norm2_lower


    def _init_layout (self):

        self._norm2_upper = None                                # cached value of norm2 deviation 
        self._norm2_lower = None                                # cached value of norm2 deviation 
        self._target_curv_le = None 
        self._target_curv_le_weighting = None

        l = QGridLayout()

        if self.target_airfoil is not None: 

            self._target_curv_le = self.target_airfoil.geo.curvature.best_around_le 

            r,c = 0, 0 
            Label  (l,r,c+1, get="Deviation", width=70, colSpan=2)
            r += 1
            Label  (l,r,c,   get="Upper Side")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_upper,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_upper()))
            r += 1
            Label  (l,r,c,   get="Lower Side")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_lower,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_lower()))
            l.setColumnMinimumWidth (c,80)
            l.setColumnMinimumWidth (c+2,20)

            r,c = 0, 3 
            Label (l,r,c, colSpan=3, get="LE curvature TE")
            r += 1
            FieldF (l,r,c  , get=lambda: self.curv_upper.max_xy[1], width=40, dec=0, 
                    style=lambda: Match_Bezier_Dialog.style_curv_le(self._target_curv_le, self.curv_upper))
            FieldF (l,r,c+1, get=lambda: self.curv_upper.te[1],     width=40, dec=1, 
                    style=lambda: Match_Bezier_Dialog.style_curv_te(self.max_curv_te_upper, self.curv_upper))

            r += 1
            FieldF (l,r,c  , get=lambda: self.curv_lower.max_xy[1], width=40, dec=0, 
                    style=lambda: Match_Bezier_Dialog.style_curv_le(self._target_curv_le, self.curv_lower))
            FieldF (l,r,c+1, get=lambda: self.curv_lower.te[1],     width=40, dec=1, 
                    style=lambda: Match_Bezier_Dialog.style_curv_te(self.max_curv_te_lower, self.curv_lower))
            l.setColumnMinimumWidth (c+2,20)

            r,c = 0, 6 
            r += 1
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.upper, self.target_upper, 
                                                            self.target_curv_le, self.max_curv_te_upper))
            r += 1
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.lower, self.target_lower, 
                                                            self.target_curv_le, self.max_curv_te_lower))
            c = 0 
            r += 1
            SpaceR (l,r, height=5, stretch=2)
            r += 1
            Label  (l,r,0, get=self._messageText, colSpan=7, height=(40, None), style=style.COMMENT)
        return l
 

    def _match_bezier (self, aSide : Side_Airfoil_Bezier, aTarget_line : Line, 
                            target_curv_le: float, max_curv_te : float  ): 
        """ run match bezier (dialog) """ 

        diag = Match_Bezier_Dialog (self, self.app_model,
                                    aSide, aTarget_line,
                                    target_curv_le = target_curv_le,
                                    max_curv_te = max_curv_te,
                                    parentPos=(0.1, 0.05), dialogPos=(0.5,1))
        diag.exec()

        if diag.made_match:

            geo : Geometry_Bezier = self.geo
            geo.finished_change_of (aSide)              # will reset and handle changed  

            self.app_model.notify_airfoil_changed()     # notify change of airfoil - new design
       

    @override
    def refresh (self, reinit_layout=False):

        # reset cached deviations
        self._norm2_lower = None
        self._norm2_upper = None 
        super().refresh(reinit_layout=reinit_layout)
        

    def _messageText (self): 
        """ user warnings"""
        text = []
        r_upper_dev = Matcher.result_deviation (self.norm2_upper())
        r_lower_dev = Matcher.result_deviation (self.norm2_lower())

        r_upper_le = Matcher.result_curv_le (self._target_curv_le, self.curv_upper)
        r_lower_le = Matcher.result_curv_le (self._target_curv_le, self.curv_lower)
        r_upper_te = Matcher.result_curv_te (self.max_curv_te_upper,self.curv_upper)
        r_lower_te = Matcher.result_curv_te (self.max_curv_te_lower, self.curv_lower)

        is_bad = Matcher.result_quality.BAD
        if r_upper_dev == is_bad or r_lower_dev == is_bad:
           text.append("- Deviation is quite high")
        if r_upper_le == is_bad or r_lower_le == is_bad:
           text.append(f"- Curvature at LE differs too much from target ({int(self._target_curv_le)})")
        if r_upper_te == is_bad or r_lower_te == is_bad:
           text.append("- Curvature at TE is quite high")

        text = '\n'.join(text)
        return text 



class Panel_Bezier_Match_Small (Panel_Bezier_Match):
    """ Match Bezier functions - small version """

    _main_margins = Panel_Airfoil_Abstract.MAIN_MARGINS_MINI

    def _init_layout (self):

        self._norm2_upper = None                                # cached value of norm2 deviation 
        self._norm2_lower = None                                # cached value of norm2 deviation 
        self._target_curv_le = None 
        self._target_curv_le_weighting = None

        l = QGridLayout()

        if self.target_airfoil is not None: 

            self._target_curv_le = self.target_airfoil.geo.curvature.best_around_le 

            r,c = 0, 0 
            Label  (l,r,c,   get="Deviation Upper")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_upper,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_upper()))
            r += 1
            Label  (l,r,c,   get="Deviation Lower")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_lower,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_lower()))
            l.setColumnMinimumWidth (c,90)
            l.setColumnMinimumWidth (c+2,20)
            r,c = 0, 3
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.upper, self.target_upper, 
                                                            self.target_curv_le, self.max_curv_te_upper))
            r += 1
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.lower, self.target_lower, 
                                                            self.target_curv_le, self.max_curv_te_lower))
            return l



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Common Utility functions for convenience - no dependencies from other modules  
"""

import os
import json
from pathlib                import Path
from typing                 import override
from termcolor              import colored
from collections.abc        import Iterable

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)


#------------------------------------------------------------------------------
# base type utils  
#------------------------------------------------------------------------------


def clip(val, min_, max_):
    """ clip aVal to be between min and max"""
    return min_ if val < min_ else max_ if val > max_ else val



#------------------------------------------------------------------------------
# logging 
#------------------------------------------------------------------------------

def init_logging (level= logging.WARNING):
    """ initialize logging with level"""

    ch = logging.StreamHandler()

    ch.setFormatter(CustomFormatter())

    logging.basicConfig(format='%(levelname)-8s- %(message)s', 
                        handlers=[ch], 
                        level=level)  # DEBUG or WARNING
    # suppress debug messages from these modules 
    # logging.getLogger('PIL.PngImagePlugin').disabled = True
    # logging.getLogger('dxf_utils').disabled = True


class CustomFormatter(logging.Formatter):
    """ colored formatting of logging stream"""

    # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG:    colored(" - %(levelname)s - %(message)s", 'yellow', attrs=["dark"]) + colored(" (%(filename)s:%(lineno)d)", 'white', attrs=["dark"]),                          # grey + format + reset,
        logging.INFO:     colored(" - %(message)s", 'white', attrs=["dark"]),
        logging.WARNING:  colored("WARNING - ", 'yellow') + "%(message)s",
        logging.ERROR:    colored("ERROR - ", 'red') + "%(message)s",
        logging.CRITICAL: colored("ERROR - ", 'red', attrs=["bold"]) + "%(message)s"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
    

#------------------------------------------------------------------------------
# Dictionary handling
#------------------------------------------------------------------------------

def fromDict(aDict : dict, key, default='no default'):
    """
    returns a value from dict. If ky is not in the dict and there is no default value an error
    will be raised 

    Args:
        dict: the dictionary to look in 
        key: the key to look for       
        default: the value if key is missing
    """
    preferred_type = None

    if default != 'no default':
        if isinstance (default, float):
           preferred_type = float
        elif isinstance (default, bool):
            preferred_type = bool
        elif isinstance (default, int):
            preferred_type = int

    if aDict and key in aDict:
        value = aDict[key]
        if preferred_type == float:
            value = float(value)
        elif preferred_type == int:
            value = int(value)
        elif preferred_type == bool:
            value = bool(value)
    else:
        if default == 'no default':
            value = None
            logger.error ('Mandatory parameter \'%s\' not specified'  % key)
        else:
            value = default 
            if value:
                logger.debug ('Parameter \'%s\' not specified, using default-value \'%s\'' % (key, str(value)))
    return value


def toDict(aDict : dict, key, value):
    """
    writes t0 the parameter dictionary. If 'value' is None the key is not written 
    """
    if isinstance (aDict,dict):
        if not value is None: 
            # limit decimals in file 
            if isinstance  (value, float):
                value = round (value,6)
            aDict [key] = value
        else: 
            # remove key from dictionary  - so default values will be used 
            aDict.pop(key, None)

        
#------------------------------------------------------------------------------
# Settings and Parameter file 
#------------------------------------------------------------------------------


class Parameters (dict):
    """ Handles a parameter file with a json structure representing a dictionary of parameters""" 

    def __init__ (self, pathFileName : str = None):
        super().__init__()

        self._pathFileName = pathFileName

        if pathFileName is not None:
            self.load()


    def __repr__ (self) -> str:
        """ nice representation of self """
        return f"<{type(self).__name__}>"


    @property
    def pathFileName (self) -> str:
        """ returns the path file name of the parameter file """
        return self._pathFileName

    def set_pathFileName (self, pathFileName : str):
        """ sets the path file name of the parameter file """
        self._pathFileName = pathFileName
        

    @override
    def get(self, key, default=None):
        """
        Returns value of 'key' 
        If a default is given, the returned value will be cast to the type of default
        """

        val = super().get(key, default)

        if default is not None and val is not None:
            if isinstance (default, float):
                val = float(val)
            elif isinstance (default, bool):
                val = bool(val)
            elif isinstance (default, int):
                val = int(val)
            elif isinstance (default, str):
                val = str(val)
        return val


    def set(self, key, value):
        """
        Sets 'key' with 'value'. If value is ... 
        - float, it will be rounded to 6 decimals
        - None or [] or {} or '' the key will be removed from the dictionary 
        """
        # avoid empty entries in settings
        if value is None or (isinstance(value, Iterable) and len(value) == 0):
            self.pop(key, None)
        else:
            # limit decimals in file
            if isinstance(value, float):
                value = round(value, 6)

            self[key] = value


    def delete (self, key, purge: bool = True):
        """ deletes 'key' from self - optionally saves self after deletion"""
        self.pop(key, None)
        if purge:
            self.save()


    def load (self):
        """
        load self from json file
        """
        dataDict = {}
        if self.pathFileName:
            try:
                fs = open(self.pathFileName)
                try:
                    dataDict = json.load(fs)
                    fs.close()
                except ValueError as e:
                    logger.error (f"Invalid json expression '{e}' in parameter file '{self.pathFileName}'")
                    fs.close()
                    dataDict = {}
            except:
                logger.debug (f"Parameter file {self.pathFileName} not found")

        self.clear()
        self.update(dataDict)


    def replace_all (self, newParams : dict):
        """ replaces all parameters of self with newParams """
        self.clear()
        self.update(newParams)


    def save (self):
        """ writes self to json file"""

        with open(self.pathFileName, 'w+') as fs:

            try:
                json.dump(self, fs, indent=2, separators=(',', ':'))
                logger.debug (f"{type(self).__name__} saved to {self.pathFileName}" )

            except ValueError as e:
                logger.error (f"Invalid json expression '{e}'. Failed to save data to '{self.pathFileName}'")
            except Exception as e:
                logger.error (f"{e}. Failed to save data to '{self.pathFileName}'")

        return


    def delete_file (self):
        """ deletes the parameter file of self """
        try:
            os.remove (self.pathFileName)
            self.clear()
            logger.debug (f"Parameter file '{self.pathFileName}' deleted")
        except Exception as e:
            logger.error (f"Failed to delete parameter file '{self.pathFileName}': {e}")



#------------------------------------------------------------------------------
# File, Path handling 
#------------------------------------------------------------------------------

class PathHandler(): 
    """ handles relative Path of actual files to a workingDir """

    def __init__ (self, workingDir=None, onFile=None): 
        """  PathHandler for working directory either from 'workingDir' directly or based 'onFile' """
        self._workingDir = None

        if workingDir is not None: 
           self.workingDir = workingDir 
        elif onFile is not None:
           self.set_workingDirFromFile (onFile)

    @classmethod
    def relPath(cls, pathFilename, start = None):
        """Return a relative version of a path - like os.path.relpath - but checks for same drive

        Args:
            :start: start dir - default: cwd (current working Dir)
        """
        if start is None: start = os.getcwd()

        if Path(pathFilename).anchor == Path(start).anchor: 
            relPath = os.path.relpath(pathFilename, start = start)
        else: 
            relPath = pathFilename
        return relPath  


    @property 
    def workingDir (self):
        return self._workingDir if (not self._workingDir is None) else ''
    
    @property 
    def workingDir_name (self): 
        """ the directory name of workingDir """
        return os.path.basename(os.path.normpath(self.workingDir))
    
    @workingDir.setter
    def workingDir (self, newDir):

        if newDir is None: 
            self._workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        elif not newDir: 
            self._workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        elif not os.path.isdir(newDir): 
            os.makedirs(newDir)  
            self._workingDir = os.path.normpath(newDir)    
        else: 
            self._workingDir = os.path.normpath(newDir) 

    def set_workingDirFromFile (self, aFilePath):

        if aFilePath is None: 
            self.workingDir = None
        elif not aFilePath: 
            self.workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        else: 
            self.workingDir = os.path.dirname(aFilePath) 
            if not os.path.isdir(self.workingDir):
                os.makedirs(self.workingDir)

    def relFilePath (self, aFilePath):
        """ returns the relative path of aFilePath to the workingDir of self"""
        if aFilePath is None: 
            return None
        else: 
            try: 
                relPath =  os.path.normpath(os.path.relpath(aFilePath, start = self.workingDir))
                if len(relPath) > len(aFilePath): 
                    return aFilePath                # relPath would be more complicated
                else: 
                    return relPath 
            except:                                 # aFilePath is on different drive 
                return aFilePath 
    
    def fullFilePath (self, aRelPath):
        """ returns the full path of relative aRelPath and the workingDir of self"""

        if aRelPath is None: 
            return self.workingDir
        else: 
            if os.path.isabs (aRelPath): 
                # maybe we can make a rel path out of it? 
                newPath = self.relFilePath (aRelPath)
                if os.path.isabs (newPath):
                    return aRelPath                 # we surrender - it's absolute
                else: 
                    aRelPath = newPath              # now we have a real rel path 
            return os.path.normpath(os.path.join (self.workingDir, aRelPath))


